#include <cstdio>
#include <list>
#include <vector>
#include <fstream>
#include <string>
#include <cmath>
#include <cfloat>
#include <GL/glut.h>

using namespace std;

class Vertex;
class Halfedge;
class Face;
class Model;

class Vec3 {
public:
	double x, y, z;
	Vec3( void );
	Vec3( const double& _x, const double& _y, const double& _z );
	Vec3& operator = ( const Vec3& );
	Vec3& operator+=( const Vec3& );
	Vec3& operator-=( const Vec3& );
  
	double length();
	double length2();
	double angle(const Vec3& v);
	void set( const double& _x, const double& _y, const double& _z );
	void normalize();
	void debugout();
	Vec3(const Vec3& v);
	void set(const Vec3& v);
	void sub( const Vec3& v0, const Vec3& v1);
	void sub( const Vec3& v);
	void add( const Vec3& v);
	void scale( const double s);
	void cross( const Vec3& v0, const Vec3& v1);
	double dot( const Vec3& v);
};

inline Vec3::Vec3( const Vec3& v ) { x = v.x; y = v.y; z = v.z;}

inline Vec3::Vec3( void ) { x = 0; y = 0; z = 0;}
inline Vec3::Vec3( const double& _x, const double& _y, const double& _z ) { x = _x; y = _y; z = _z; }
inline Vec3& Vec3::operator =  ( const Vec3& v ){ x = v.x; y = v.y; z = v.z; return *this; }
inline Vec3 operator - ( const Vec3& v ) { return( Vec3( -v.x, -v.y, -v.z ) ); }
inline Vec3& Vec3::operator+=( const Vec3& v ) { x += v.x; y += v.y; z += v.z; return( *this ); }
inline Vec3& Vec3::operator-=( const Vec3& v ) { x -= v.x; y -= v.y; z -= v.z; return( *this ); }
inline Vec3 operator+( const Vec3& v1, const Vec3& v2 ) { return( Vec3( v1.x+v2.x, v1.y+v2.y, v1.z+v2.z ) );}
inline Vec3 operator-( const Vec3& v1, const Vec3& v2 ) { return( Vec3( v1.x-v2.x, v1.y-v2.y, v1.z-v2.z ) );}
inline Vec3 operator*( const double& k, const Vec3& v ) { return( Vec3( k*v.x, k*v.y, k*v.z ) );}
inline Vec3 operator*( const Vec3& v, const double& k ) { return( Vec3( v.x*k, v.y*k, v.z*k ) );}
inline Vec3 operator/( const Vec3& v, const double& k ) { return( Vec3( v.x/k, v.y/k, v.z/k ) );}
//----  ς̒`
inline double operator*( const Vec3& v1, const Vec3& v2 ) { return( v1.x*v2.x + v1.y*v2.y + v1.z*v2.z );}
//----  Oς̒`
inline Vec3  operator%( const Vec3& v1, const Vec3& v2 ) { return( Vec3( v1.y*v2.z - v1.z*v2.y,  v1.z*v2.x - v1.x*v2.z,  v1.x*v2.y - v1.y*v2.x ) );}
inline double Vec3::length() { return( sqrt( x*x + y*y + z*z ) );}
inline double Vec3::length2() { return( x*x + y*y + z*z  );}
inline void Vec3::normalize() { double l = length(); x/=l; y/=l; z/=l;}
inline void Vec3::debugout() { printf("Vec3(%f %f %f)\n", x, y, z );}
inline void Vec3::set( const double& _x, const double& _y, const double& _z ) { x = _x; y = _y; z = _z; }
inline double Vec3::angle(const Vec3& v) {
        Vec3 c = (*this)%v;
        return atan2(c.length(), (*this)*v);
}

class Vertex {
public:
	Vec3 p;
	Halfedge *halfedge;

	Vertex(double _x, double _y, double _z) {
		p.x = _x;
		p.y = _y;
		p.z = _z;
		halfedge = NULL;
	}
};

class Halfedge {
public:
	Vertex *vertex;
	Face *face;
	Halfedge *pair;
	Halfedge *next;
	Halfedge *prev;

	Halfedge(Vertex *v) {
		vertex = v;
		if(v->halfedge == NULL) {
			v->halfedge = this;
		}
	}
};

class Face {
public:
	Halfedge *halfedge;
	Vec3 normal;

	Face(Halfedge *he) {
		halfedge = he;
	}

	void setNormal() {
		Vec3 p0 = halfedge->vertex->p;
		Vec3 p1 = halfedge->next->vertex->p;
		Vec3 p2 = halfedge->next->next->vertex->p;

		normal = (p1 - p0) % (p2 - p0);
		normal.normalize();
	}

};

class Model {
public:
	std::list<Face*> faces;
	std::list<Vertex*> vertices;

private:
	void setHalfedgePair( Halfedge *he ) {
		std::list<Face*>::iterator it_f;
		for( it_f = faces.begin(); it_f != faces.end(); it_f++ ) {
			Halfedge *halfedge_in_face = (*it_f)->halfedge;
			do {
				if(he->vertex == halfedge_in_face->next->vertex &&
					he->next->vertex == halfedge_in_face->vertex) {
					he->pair = halfedge_in_face;
					halfedge_in_face->pair = he;
					return;
				}
				halfedge_in_face = halfedge_in_face->next;
			} while(halfedge_in_face != (*it_f)->halfedge);
		}	
	}

	void setHalfedgePair(Halfedge *he0, Halfedge *he1) {
		he0->pair = he1;
		he1->pair = he0;
	}

public:
	void sizeNormalize() {
		std::list<Vertex*>::iterator it_v;
		Vec3 minV(DBL_MAX, DBL_MAX, DBL_MAX);
		Vec3 maxV(-DBL_MAX, -DBL_MAX, -DBL_MAX);
		for( it_v = vertices.begin(); it_v != vertices.end(); it_v++ ) {
			minV.x = min(minV.x, (*it_v)->p.x);
			minV.y = min(minV.y, (*it_v)->p.y);
			minV.z = min(minV.z, (*it_v)->p.z);
			maxV.x = max(maxV.x, (*it_v)->p.x);
			maxV.y = max(maxV.y, (*it_v)->p.y);
			maxV.z = max(maxV.z, (*it_v)->p.z);
		}

		Vec3 center = 0.5*(minV + maxV);
		double size = (maxV - minV).length();
		for( it_v = vertices.begin(); it_v != vertices.end(); it_v++ ) {
			(*it_v)->p = ((*it_v)->p - center) / size / 2;
		}
	}

	void addFace( Vertex *v0, Vertex *v1, Vertex *v2 ) {
		Halfedge* he0 = new Halfedge( v0 );
		Halfedge* he1 = new Halfedge( v1 );
		Halfedge* he2 = new Halfedge( v2 );

		he0->next = he1;	he0->prev = he2;
		he1->next = he2;	he1->prev = he0;
		he2->next = he0;	he2->prev = he1;

		Face *face = new Face( he0 );
		faces.push_back( face );
		he0->face = face;
		he1->face = face;
		he2->face = face;
		setHalfedgePair( he0 );
		setHalfedgePair( he1 );	
		setHalfedgePair( he2 );
		face->setNormal();
	}

};

	
Model *g_model;

Model* readFile(const char* filename) {

	std::ifstream ifs(filename);
	std::string token(BUFSIZ, '\0');

	Model *model = new Model();
	std::vector<Vertex*> vertices; // CfbNXŒ_QƂ邽߂vector

	while (!(ifs >> token).fail()) {
		if (token[0] == '#') {
			ifs.ignore(BUFSIZ, '\n');
			continue;
		}

		if (token == "v") {
			double x, y, z;
			ifs >> x >> y >> z;
			ifs.ignore(BUFSIZ, '\n');

			Vertex *v = new Vertex(x, y, z);
			model->vertices.push_back(v);
			vertices.push_back(v);	
		}
		else if (token == "f"){
			int index0, index1, index2;
			ifs >> index0 >> index1 >> index2;
			ifs.ignore(BUFSIZ, '\n');
		
			model->addFace( vertices[index0-1], vertices[index1-1], vertices[index2-1] );

		} else {
			ifs.ignore(BUFSIZ, '\n'); 
		}
	}

	model->sizeNormalize();
	return model;
}


double rotateAngleH_deg; // ʐ̉]px
double rotateAngleV_deg; // c̉]px
int preMousePositionX;   // }EXJ[\̈ʒuLĂϐ
int preMousePositionY;   // }EXJ[\̈ʒuLĂϐ
bool mouseLeftPressed;
bool mouseRightPressed;
double modelScale;

const GLfloat lightDiffuse[] = { 1 , 1 , 1 , 1 };
const GLfloat materialDiffuseFront[] = { 0 , 1.0 , 0 , 1};
const GLfloat materialDiffuseBack[] = { 1.0 , 1.0 , 0 , 1};
const GLfloat lightPosition[] = { 50 , 50 , 50 , 0 };

void drawModel() {
	std::list<Vertex*>::iterator it_v;

	// _̕`
/*	
	glDisable(GL_LIGHTING);
	glColor3d(0,0,0);
	glPointSize(5);
	glBegin(GL_POINTS);
	for( it_v = g_model->vertices.begin(); it_v != g_model->vertices.end(); it_v++ ) {
		glVertex3d((*it_v)->p.x, (*it_v)->p.y, (*it_v)->p.z);
	}
	glEnd();
*/
	// Ő̕`
	glDisable(GL_LIGHTING);
	glColor3d(0,0,0);
	std::list<Face*>::iterator it_f;
	for( it_f = g_model->faces.begin(); it_f != g_model->faces.end(); it_f++ ) {
		glBegin(GL_LINE_LOOP);
		Halfedge *he = (*it_f)->halfedge;
		do {
			glVertex3d(he->vertex->p.x, he->vertex->p.y, he->vertex->p.z);
			he = he->next;
		} while(he != (*it_f)->halfedge);
		glEnd();
	}


	// ʂ̕`
	glEnable(GL_LIGHTING);

	// \ʂ̕`
    glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, materialDiffuseFront);

	for( it_f = g_model->faces.begin(); it_f != g_model->faces.end(); it_f++ ) {
		glBegin(GL_POLYGON);
		glNormal3f( (*it_f)->normal.x, (*it_f)->normal.y, (*it_f)->normal.z);

		Halfedge *he = (*it_f)->halfedge;
		do {
			glVertex3d(he->vertex->p.x, he->vertex->p.y, he->vertex->p.z);
			he = he->next;
		} while(he != (*it_f)->halfedge);
		glEnd();
	}

	// ʂ̕`
    glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, materialDiffuseBack);
	for( it_f = g_model->faces.begin(); it_f != g_model->faces.end(); it_f++ ) {
		glBegin(GL_POLYGON);
		glNormal3f( -(*it_f)->normal.x, -(*it_f)->normal.y, -(*it_f)->normal.z);

		Halfedge *he = (*it_f)->halfedge;
		do {
			glVertex3d(he->vertex->p.x, he->vertex->p.y, he->vertex->p.z);
			he = he->prev;
		} while(he != (*it_f)->halfedge);
		glEnd();
	}


}

void display(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslated(0, 0.0, -1.0 / modelScale);

	glLightfv(GL_LIGHT0 , GL_POSITION , lightPosition);

	glRotated(rotateAngleV_deg, 1.0, 0.0, 0.0);
	glRotated(rotateAngleH_deg, 0.0, 1.0, 0.0);

	drawModel();

	glutSwapBuffers();
}

void resize(int w, int h) {
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(30.0, (double)w / (double)h, 0.1, 100.0);
	glMatrixMode(GL_MODELVIEW);

}

void keyboard(unsigned char key, int x, int y) {
	if (key == 'a') {
	}
}

void mouse(int button, int state, int x, int y) {
	mouseLeftPressed = mouseRightPressed = false;

	switch (button) {
	case GLUT_LEFT_BUTTON:
		preMousePositionX = x;
		preMousePositionY = y;
		mouseLeftPressed = true;
		break;
	case GLUT_MIDDLE_BUTTON:
		break;
	case GLUT_RIGHT_BUTTON:
		preMousePositionX = x;
		preMousePositionY = y;
		mouseRightPressed = true;
		break;
	default:
		 break;
	}
}


void motion(int x, int y) {
	int diffX = x - preMousePositionX;
	int diffY = y - preMousePositionY;

	if(mouseLeftPressed) {
		rotateAngleH_deg += diffX * 0.3;
		rotateAngleV_deg += diffY * 0.3;
	} else if(mouseRightPressed) {
		double scale = 1.0 + diffX * 0.01;
		modelScale *= scale;
	}

	preMousePositionX = x;
	preMousePositionY = y;
	glutPostRedisplay();
}


void init(void) {
	modelScale = 1.0;
	glClearColor(1.0, 1.0, 1.0, 0.0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);


	glLightfv(GL_LIGHT0 , GL_DIFFUSE , lightDiffuse);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

    glEnable (GL_POLYGON_OFFSET_FILL);
    glPolygonOffset (1.0, 0);

	glEnable(GL_CULL_FACE);

}

int main(int argc, char* argv[]) {
	if (argc != 2) {
		printf("Usage: specify an objfile\n");
		return 1;
	}

	g_model = readFile(argv[1]);

	glutInit(&argc, argv);
	glutInitWindowSize(600,600);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutReshapeFunc(resize);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	init();
	glutMainLoop();

	return 0;
}
